// we are including all the PM header files here.  Please replace with you
// stock "Include all the OS/2 PM stuff" include file
#define INCL_WIN
#define INCL_DOS
#define INCL_DOSDEVIOCTL

extern "C"
{
#include <os2.h>
}

// include the Mesa AddIn header file
#include "extaddin.h"
#include <string.h>
#include <stdio.h>                 // added for printf
#include <stdlib.h>
#include <math.h>

// to avoid the C++ name mangling, let's define the names of the external
// function here

#include "model.h"
#include "range.h"

void _System FactorialFunc(int numParam, void *stack, void *model_handle)
{
    Model *mo = new Model((int) model_handle);
    int x;

    if (numParam != 1) {  // the number of parameters is incorrect
        for (x = 0; x < numParam ; x++) { // clear the stack
            MesaAddInValue aiv;
            mo -> initAddInValue(aiv);
            mo -> popValue(stack,aiv);
            mo -> freeAddInValue(aiv);
        } // endfor
        mo->pushError(stack,21);   // push the error on the stack
        // 21 is incorrect number of parameters
    } else {
        double a;
        int error = 0;
        if (mo->popValue(stack,a)) error = 1;  // pop the duration
        if (!error) {  // if none of the parameters was in error,
            int x = a;
            double result = x;
            while (x) {
                result *= x;
                x--;
            }

            mo->pushValue(stack,result);
        } else {
            mo->pushError(stack,26); // push an error onto the stack
        } // endif
    } // endif

    delete mo;
}

void _System BeepFunction(int numParam, void *stack, void *model_handle)
{
    Model *mo = new Model((int) model_handle);
    int x;

    if (numParam != 2) {  // the number of parameters is incorrect
        for (x = 0; x < numParam ; x++) { // clear the stack
            MesaAddInValue aiv;
            mo -> initAddInValue(aiv);
            mo -> popValue(stack,aiv);
            mo -> freeAddInValue(aiv);
        } // endfor
        mo->pushError(stack,21);   // push the error on the stack
        // 21 is incorrect number of parameters
    } else {
        double a,b;
        int error = 0;
        if (mo->popValue(stack,a)) error = 1;  // pop the duration
        if (mo->popValue(stack,b)) error = 1;  // pop the frequency
        if (!error) {  // if none of the parameters was in error,
            int d = a,f = b;      // Beep
            DosBeep(f,d);
            mo->pushValue(stack,"BEEP");  // We need to push something onto the stack
        } else {
            mo->pushError(stack,26); // push an error onto the stack
        } // endif
    } // endif

    delete mo;
}

void _System EngineeringFunction(int numParam, void *stack, void *model_handle)
{
    Model *mo = new Model((int) model_handle);
    int x;

    if (numParam != 1 && numParam != 2) {  // the number of parameters is incorrect
        for (x = 0; x < numParam ; x++) { // clear the stack
            MesaAddInValue aiv;
            mo -> initAddInValue(aiv);
            mo -> popValue(stack,aiv);
            mo -> freeAddInValue(aiv);
        } // endfor
        mo->pushError(stack,21);   // push the error on the stack
        // 21 is incorrect number of parameters
    } else {
        double a,b;
        int error = 0;
        b = 3.0;
        if (mo->popValue(stack,a)) error = 1;  // pop the duration
        if (numParam == 2) {
            b = a;
            if (mo->popValue(stack,a)) error = 1;  // pop the duration
        } /* endif */

        if (!error) {  // if none of the parameters was in error,
            char buf[100];
            char buf2[10];
            int sgn = (a >= 0.0)?1:-1;
            a = fabs(a);
            int e = 0;
            e = log10(a);
            int sgne = (e >= 0.0)?1:-1;
            a /= pow(10,e);
            e = abs(e);
            while (e % 3) {
                if (sgne == 1) {
                    a *= 10;
                } else {
                    a /= 10;
                } /* endif */
                e--;
            } /* endwhile */
            a *= sgn;
            e *= sgne;
            sprintf(buf,"%.*lfe%d",(int)b,a,e);
            mo->pushValue(stack,buf);
        } else {
            mo->pushError(stack,26); // push an error onto the stack
        } // endif
    } // endif

    delete mo;
}

void _System numIn(int numParam, void *stack, void *model_handle)
{
    Model *mo = new Model((int) model_handle);
    MesaAddInValue aiv, aiv2;
    MesaAddInValue *ap;
    int count = 0;


    if (numParam != 2) { // incorrect number of parameters
        for (int x = 0; x < numParam ; x++) { // clear the stack
            mo -> initAddInValue(aiv);
            mo -> popValue(stack,aiv);
            mo -> freeAddInValue(aiv);
        } // endfor
        mo->pushError(stack,21);   // push the error on the stack
        delete mo;
        return;
    } // endif

    mo -> initAddInValue(aiv);
    mo -> initAddInValue(aiv2);

    mo -> popValue(stack,aiv);  // get the array
    mo -> popValue(stack,aiv2); // get the search value

    ap = aiv.array;

    for (int r = 0; r < aiv.rows ; r++) {  // increment throught the rows and columns
        for (int c = 0; c < aiv.cols ; c++) {
            if (ap->type == aiv2.type) {
                switch (ap->type) {
                case errorValueMesaAddInType:
                    if (ap->error == aiv2.error) {
                        count++;
                    } // endif
                    break;
                case stringValueMesaAddInType:
                    if (!strcmp(ap->string,aiv2.string)) {
                        count++;
                    } // endif
                    break;
                case numberValueMesaAddInType:
                    if (ap->number == aiv2.number) {
                        count++;
                    } // endif
                    break;
                } // endswitch
            } // endif
            ap++;
        } // endfor
    } // endfor


    mo -> freeAddInValue(aiv);
    mo -> freeAddInValue(aiv2);

    double result = count;
    mo -> pushValue(stack,result); // push the rsult onto the stack
    delete mo;
}

void _System layername(int numParam, void *stack, void *model_handle)
{
    Model *mo = new Model((int) model_handle);
    int x;
    double d;
    char buffer[1000];

    if (numParam != 1) {  // this only takes one parameter
        for (x = 0; x < numParam ; x++) { // clear the stack
            MesaAddInValue aiv;
            mo -> initAddInValue(aiv);
            mo -> popValue(stack,aiv);
            mo -> freeAddInValue(aiv);
        } // endfor
        mo -> pushError(stack,21);   // push the error on
        // 21 is incorrect number of parameters
    } else {
        double a,b;
        int error = 0;
        if (mo -> popValue(stack,d)) error = 1;  // pop a value off
        // if it isn;t a number, mthe parameter is wrong
        if (!error) {
            int l = d;
            char * c = mo -> getLayerName(l);
            mo -> pushValue(stack,c); // push it on the stack
            SysMFree(c);
        } else {
            mo -> pushError(stack,26);
        } // endif
    } // endif

    delete mo;
}

