
#define INCL_WIN
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "range.h"
#include "model.h"
#include "addin.h"
#include "extaddin.h"
#include "solver.h"
#include "sampadin.h"

extern HWND *theSolverHwnds;
extern AddIn *thisAddIn;
extern int solvCount;


// the data needed in the goal seeker.
struct _solveData {
    int type;
    Model *md;
    Solver *sp;
    ULONG idTimer;
};

// this is the procedure that processes the "goal seeker" Dialog
MRESULT EXPENTRY SolvDlgProc(HWND hwnd,ULONG msg,MPARAM mp1,MPARAM mp2)
{
    MRESULT ret = (MRESULT) 0L;
    _solveData *sd = (_solveData *)WinQueryWindowULong(hwnd,QWL_USER);
    char buffer[100];
    int tmp, x;
    Address ad;

    switch (msg) {
    case WM_INITDLG:
        x = 0;  // add itself to the array of hwnds
        while (x < solvCount && theSolverHwnds[x]) {
            x++;
        } /* endwhile */

        if (x == solvCount) {
            solvCount *= 2;
            theSolverHwnds = (HWND *)MRealloc(theSolverHwnds,solvCount*sizeof(HWND));
        } /* endif */

        theSolverHwnds[x] = hwnd;

        sd = new _solveData;
        WinSetWindowULong(hwnd,QWL_USER,(ULONG) sd);
        sd->md = new Model(-1);
        sd->sp = NULL;

        ad=thisAddIn->getCurrentAddress();

        if (! sd->md->getBlob("SOLVER_CHNGCELL",100,buffer))
            sd->md->addressToString(ad,buffer,100);
        WinSetDlgItemText(hwnd,DID_CHANGECELL,buffer);

        if (! sd->md->getBlob("SOLVER_FINALCELL",100,buffer))
            sd->md->addressToString(ad,buffer,100);
        WinSetDlgItemText(hwnd,DID_SETCELL,buffer);

        if (! sd->md->getBlob("SOLVER_PREC",100,buffer))
            strcpy(buffer,"0.0001");
        WinSetDlgItemText(hwnd,DID_PRECISION,buffer);

        if (! sd->md->getBlob("SOLVER_FVAL",100,buffer))
            strcpy(buffer,"0.0");
        WinSetDlgItemText(hwnd,DID_FINALVALUE,buffer);

        if (! sd->md->getBlob("SOLVER_TYPE",100,buffer))
            sd->type = 0;
        else
            sd->type = atoi(buffer);

        if (sd->type < 0 || sd->type > 2) {
            sd->type = 0;
        } /* endif */

        WinSendDlgItemMsg(hwnd,sd->type+DID_VALUE,BM_SETCHECK,(PVOID)TRUE,0L);

        sd->md->getName(buffer,100);
        WinSetDlgItemText(hwnd,DID_MODELNAME,buffer);

        WinEnableWindow(WinWindowFromID(hwnd,DID_STOP),FALSE);
        WinEnableWindow(WinWindowFromID(hwnd,DID_START),TRUE);
        break;
    case WM_CONTROL:
        switch (SHORT1FROMMP(mp1)) {
        case DID_MINIMUM:
            sd->type = 1;
            break;
        case DID_MAXIMUM:
            sd->type = 2;
            break;
        case DID_VALUE:
            sd->type = 0;
            break;
        default:
            return WinDefDlgProc ( hwnd,msg,mp1,mp2 ) ;
        } // end switch
        break;
    case WM_COMMAND:
        switch (LOUSHORT(mp1)) {
        case DID_START:
            {
                Address adc,ads;
                double prec,fv;
                int type;

                WinQueryDlgItemText(hwnd,DID_CHANGECELL,100,buffer);
                sd->md->setBlob("SOLVER_CHNGCELL",strlen(buffer)+1,buffer);
                adc = sd->md->stringToAddress(buffer);
                WinQueryDlgItemText(hwnd,DID_SETCELL,100,buffer);
                sd->md->setBlob("SOLVER_FINALCELL",strlen(buffer)+1,buffer);
                ads = sd->md->stringToAddress(buffer);

                WinQueryDlgItemText(hwnd,DID_FINALVALUE,100,buffer);
                sd->md->setBlob("SOLVER_FVAL",strlen(buffer)+1,buffer);
                fv = strtod(buffer,NULL);
                WinQueryDlgItemText(hwnd,DID_PRECISION,100,buffer);
                sd->md->setBlob("SOLVER_PREC",strlen(buffer)+1,buffer);
                prec = strtod(buffer,NULL);

                sprintf(buffer,"%d",sd->type);
                sd->md->setBlob("SOLVER_TYPE",strlen(buffer)+1,buffer);

                sd->sp = new Solver(*sd->md,adc,ads,prec,sd->type,fv);

                WinEnableWindow(WinWindowFromID(hwnd,DID_START),FALSE);
                WinEnableWindow(WinWindowFromID(hwnd,DID_STOP),TRUE);

                sd->idTimer = WinStartTimer(WinQueryAnchorBlock(hwnd),hwnd,TID_USERMAX - 50,500);
            }
            break;
        case DID_STOP:
            WinStopTimer(WinQueryAnchorBlock(hwnd),hwnd,TID_USERMAX - 50);
            if (sd->sp) {
                if (sd->sp->isError()) {
                    sd->md->redisplay();
                    WinMessageBox(HWND_DESKTOP,HWND_DESKTOP,sd->sp->getError(),"Goal Seeker Error!",
                                  24L,MB_OK | MB_ERROR);
                } else if (sd->sp->percentComplete() > 99) {
                    DosBeep(1440,100);
                    sd->md->redisplay();
                } else {
                    sd->sp->stop();
                    sd->md->redisplay();
                    WinMessageBox(HWND_DESKTOP,HWND_DESKTOP,sd->sp->getError(),"Goal Seeker Error!",
                                  24L,MB_OK | MB_ERROR);
                }
                delete sd->sp;
                sd->sp = NULL;
            }
            WinEnableWindow(WinWindowFromID(hwnd,DID_STOP),FALSE);
            WinEnableWindow(WinWindowFromID(hwnd,DID_START),TRUE);
            break;
        default:
            ret = WinDefDlgProc(hwnd,msg,mp1,mp2);
            break;
        } /* endswitch */
        break;
    case WM_TIMER:
        if ((ULONG)mp1 == sd->idTimer) {
            if (sd->sp && (sd->sp->isError() || (sd->sp->percentComplete() > 99 ))) {
                WinStopTimer(WinQueryAnchorBlock(hwnd),hwnd,TID_USERMAX - 50);
                sd -> idTimer = 0;
                WinPostMsg(hwnd,WM_COMMAND,MPFROMSHORT(DID_STOP),0L);
            } else {
                performAddInFunction(cycleCalcIndicatorMesaAddIn);
            } /* endif */
        } else {
            WinDefDlgProc(hwnd,msg,mp1,mp2);
        } /* endif */
        break;
    case WM_GETMODELPTR:
        return (MPARAM)sd->md;
        break;
    case WM_CLOSE:
        for (x = 0; x < solvCount ; x++) {   // remove from the hwnd Array
            if (theSolverHwnds[x] == hwnd) {
                theSolverHwnds[x] = NULL;
            } /* endif */
        } /* endfor */

        if (sd->sp) {
            delete sd->sp;
        } /* endif */
        if (sd->md) {
            delete sd->md;
        } /* endif */
        delete sd;
        WinDismissDlg(hwnd,DID_OK);
        break;
    default:
        ret = WinDefDlgProc(hwnd,msg,mp1,mp2);
        break;
    } /* endswitch */

    return ret;
}



double waitForNewValue(Solver &sv, Model &model, Address &ad, double oldval, int &maxtime)
{
    int totaltime = 1;
    double newValue;
    model.getCellValue(ad,newValue);

    while (( totaltime < maxtime)  && (newValue == oldval)) {
         DosSleep(50);
         totaltime += 50;
         model.getCellValue(ad,newValue);
         if (sv.shouldStop()) {
             DosExit(0,0);  //kill the thread
         } // endif
    }

    if (totaltime < maxtime) {
       maxtime = 0;
    }

    return newValue;
}

double func(Solver & sv,Model &model, Address & x, Address & y, double newX)
{
    double oldX;
    double oldY;
    double newY;
    int maxtime = 30000;  // 30 second wait time

    model.getCellValue(x,oldX);
    model.getCellValue(y,oldY);

    if (oldX == newX) return oldY;   // the new X value is the same as the old one

    model.setAction(0x10);
    model.setCellValue(x,newX);
    model.setAction(0);
    newY = waitForNewValue(sv,model,y,oldY,maxtime);
    return newY;
}

int bracketRoot(Solver & sv, Model &model, Address & x, Address & y, double &lft, double &rght, double offset)
{
    int j;
    double f1,f2;
    double gap;

    if (lft > rght) {
       f1 = lft;
       lft = rght;
       rght = f1;
    }
    if (lft == rght) rght += 0.1;
    gap = rght - lft;
    f1 = func(sv,model,x,y,lft) - offset;
    f2 = func(sv,model,x,y,rght) - offset;
    for (j = 0 ; j < 100 ; j++) {
         if (f1 * f2 < 0.0) return 1;
         if (fabs(f1) < fabs(f2)) {
             gap *= 2;
             rght = lft;
             lft -= gap;
             f2 = f1;
             f1 = func(sv,model,x,y,lft) - offset;
         } else {
             gap *=2;
             lft = rght;
             rght += gap;
             f1 = f2;
             f2 = func(sv,model,x,y,rght) - offset;
         }
    }
    return 0;
}

int bisectToRoot(Solver & sv,Model &model,Address & x, Address & y, double &lft, double &rght, double offset, double acc)
{
   int j;
   double dx,f,fmid,xmid,rtb;

    f = func(sv,model,x,y,lft) - offset;
    dx = rght - lft;
    xmid = (rght+lft)/2;   // find midpoint;
    for ( j = 0 ; j < 40 ; j ++ ) {
         fmid=func(sv,model,x,y,xmid) - offset;
         if (((fmid < 0.0)  && (f < 0.0)) || ((fmid > 0.0)  && (f > 0.0))){
            lft = xmid;
            f = fmid;
         } else {
            rght = xmid;
         }
         dx = rght - lft;
         if (fabs(dx) < acc || fmid == 0) {
             lft=rght=xmid;
             return 1;
         }
         xmid = (rght+lft)/2;   // find midpoint;
    }
    return 0;
}

void _System solveThread(unsigned long p)
{
    Solver * sp = (Solver *)p;
    double left,right;

    sp->model().getCellValue(sp->changeCell(),left);
    right = left;

    if (!bracketRoot(*sp,sp->model(),sp->changeCell(), sp->finalCell(), left, right, sp->finalValue())) {
       sp->setError("Could not bracket value!  Try a better guess.");
       return;
    }
    if (!bisectToRoot(*sp,sp->model(),sp->changeCell(),sp->finalCell(),
        left,right,sp->finalValue(),sp->precision())) {
        sp->setError("Could not find value! Try a better guess or lowering the precision.");
        return;
    }
    sp->setPercentComplete(100);
}



int bracketMinMax(Solver &sv,Model &model, Address & x, Address & y,
               double &lft, double &fl,double &rght, double &fr, int min)
{
    int j;
    double fmid, mid;
    double gap;

    if (lft > rght) {
       fl = lft;
       lft = rght;
       rght = fl;
    }
    if (lft == rght) rght += 0.1;
    fl = func(sv,model,x,y,lft);
    fr = func(sv,model,x,y,rght);
    mid = (rght + lft) / 2;
    gap = (rght - lft) / 2;
    fmid = func(sv,model,x,y,mid);

    for (j = 0 ; j < 100 ; j++) {
         if (( fl > fmid ) && ( fr > fmid ) && min) return 1;
         if (( fl < fmid ) && ( fr < fmid ) && !min) return 1;
         if (((fl < fr) && min) || ((fl > fr) && !min)) {
             gap *= 2;
             rght = mid;
             mid = lft;
             lft -= gap;
             fr = fmid;
             fmid = fl;
             fl = func(sv,model,x,y,lft);
         } else {
             gap *=2;
             lft = mid;
             mid = rght;
             rght += gap;
             fl = fmid;
             fmid = fr;
             fr = func(sv,model,x,y,rght);
         }
    }
    return 0;
}



int bisectToMinMax(Solver & sv,Model &model,Address & x, Address & y,
                double &lft,double & fl,double &rght,double &fr, double acc, int min)
{
    double gap = (fabs(rght - lft)) / 7;
    double dx = lft;
    double fx;
    double oldfx, oldoldfx;

    oldfx = fx = oldoldfx = fl;

    for (int j = 0; j < 40; j++ ) {
        while ((dx < rght) && (((oldfx >= fx) && min) ||((oldfx <= fx) && !min)))  {
            oldoldfx = oldfx;
            oldfx = fx;
            dx += gap;
            fx = func(sv,model,x,y,dx);
        } /* endwhile */

        rght = dx;
        fr = fx;

        dx -= 2*gap;

        if (dx < lft) {
           dx = lft;
        } /* endif */

        lft = dx;

        if (fabs(lft-rght) < acc) {
           lft = rght;
            return 1;
        } /* endif */

        fl = oldfx = fx = oldoldfx;
        gap = (fabs(rght - lft)) / 7;
    } /* endfor */
    return 0;
}


void _System minimizeThread(unsigned long p)
{
    Solver * sp = (Solver *)p;
    double left,right, fl,fr;

    sp->model().getCellValue(sp->changeCell(),left);
    right = left;

    if (!bracketMinMax(*sp,sp->model(),sp->changeCell(), sp->finalCell(), left, fl,right,fr, 1)) {
       sp->setError("Could not bracket minimum!  Try a better guess.");
       return;
    }
    if (!bisectToMinMax(*sp,sp->model(),sp->changeCell(),sp->finalCell(),
        left,fl,right,fr,sp->precision(),1)) {
        sp->setError("Could not find minimum! Try a better guess or lowering the precision.");
        return;
    }
    sp->setPercentComplete(100);

}

void _System maximizeThread(unsigned long p)
{
    Solver * sp = (Solver *)p;
    double left,right, fl,fr;

    sp->model().getCellValue(sp->changeCell(),left);
    right = left;

    if (!bracketMinMax(*sp,sp->model(),sp->changeCell(), sp->finalCell(), left, fl ,right, fr, 0)) {
       sp->setError("Could not bracket maximum!  Try a better guess.");
       return;
    }
    if (!bisectToMinMax(*sp,sp->model(),sp->changeCell(),sp->finalCell(),
        left,fl,right,fr,sp->precision(),0)) {
        sp->setError("Could not find maximum! Try a better guess or lowering the precision.");
        return;
    }
    sp->setPercentComplete(100);

}


Solver :: Solver(Model & theModel, Address & change, Address &final,
               double precision, int type, double value) :
    _model(theModel),chng(change),finl(final),prec(precision),
    tp(type),fval(value)
{
    error[0] = 0;
    percentComp = 0;
    dostop = 0;
    switch (tp) {
    case 0:
       DosCreateThread(&tid,&solveThread,(ULONG)this,0L,16384);
       break;
    case 1:
       DosCreateThread(&tid,&minimizeThread,(ULONG)this,0L,16384);
       break;
    case 2:
       DosCreateThread(&tid,&maximizeThread,(ULONG)this,0L,16384);
       break;
    } /* endswitch */
}

    // stop the solving process, blocks until the solving thread is done
Solver & Solver :: stop()
{
//    int rc = DosKillThread(tid);
      // don't use DosKillThread.  it can leave MesaCore in an unknown state
      // and possibly cause crashes
    dostop = 1;
    setError("Stopped before completion!");
    return *this;
}

Solver :: ~Solver()
{
    dostop = 1;
    DosWaitThread(&tid,DCWW_WAIT);
}





